import os
import rpm
import shutil
import re
import subprocess
import tempfile
import git
import glob

try:
    from urllib.parse import urlparse
except ImportError:
     from urlparse import urlparse

import pyrpkg
from pyrpkg.utils import cached_property
from pyrpkg.errors import rpkgError
from pyrpkg.sources import SourcesFile

from rpkglib.lookaside import CGILookasideCache
from rpkglib.preproc import preprocess
from rpkglib import utils

from .exceptions import NotUnpackedException, RpmSpecParseException, NoSourceZeroException


class Commands(pyrpkg.Commands):
    def __init__(self, *args, **kwargs):
        """Init the object and some configuration details."""
        self._outdir = kwargs.pop("outdir", None)
        self.preprocess = kwargs.pop("preprocess", True)
        self.base_output_path = kwargs.pop("base_output_path", '/tmp/rpkg')
        self.dump_spec_only = False
        self.version_bump = False
        self.use_existing_outdir = False
        super(Commands, self).__init__(*args, **kwargs)
        self.source_entry_type = 'bsd'
        self.distgit_namespaced = True
        self.lookaside_namespaced = True
        self._specloc = self.path
        self._ns_module_name = None
        self._repo_spec = None
        self._module_name_spec = None

    def set_specloc(self, specloc):
        if not specloc:
            self._specloc = self.path
        else:
            self._specloc = specloc

    def load_rpmdefines(self):
        self._rpmdefines = [
            "--define '_sourcedir %s'" % self.outdir,
            "--define '_specdir %s'" % self.outdir,
            "--define '_builddir %s'" % self.outdir,
            "--define '_buildrootdir %s'" % self.outdir,
            "--define '_srcrpmdir %s'" % self.outdir,
            "--define '_rpmdir %s'" % self.outdir,
        ]

    def load_repo(self):
        """Create a repo object from our path"""
        self.log.debug('Creating repo object from %s', self.path)
        try:
            self._repo = git.Repo(self.path, search_parent_directories=True)
        except (git.InvalidGitRepositoryError, git.NoSuchPathError):
            raise rpkgError('%s is not a valid repo' % self.path)

    @property
    def outdir(self):
        if not self._outdir:
            self.load_outdir()
        return self._outdir

    def load_outdir(self):
        dirname_base = os.path.splitext(
            os.path.splitext(os.path.basename(self.repo_spec))[0])[0]
        cnt = len(glob.glob(os.path.join(self.base_output_path, dirname_base+'*')))

        if not self.use_existing_outdir:
            prefix = dirname_base + '-' + str(cnt+1) + '-'
            self._outdir = tempfile.mkdtemp(dir=self.base_output_path, prefix=prefix)
            return

        try:
            prefix = dirname_base + '-' + str(cnt) + '-'
            self._outdir = glob.glob(os.path.join(self.base_output_path, prefix+'*'))[0]
        except IndexError:
            raise rpkgError('Could not load the latest outdir with prefix "%s" at %s'
                            % (prefix, self.base_output_path))

    @property
    def repo_spec(self):
        if not self._repo_spec:
            self.load_repo_spec()
        return self._repo_spec

    def load_nameverrel(self):
        """Get name, version, release of a package module."""

        cmd = ['rpm']
        cmd.extend(self.rpmdefines)
        # We make sure there is a space at the end of our query so that
        # we can split it later.  When there are subpackages, we get a
        # listing for each subpackage.  We only care about the first.
        cmd.extend(['-q', '--qf', '"??%{NAME} %{EPOCH} %{VERSION} %{RELEASE}??"',
                    '--specfile', '"%s"' % os.path.join(self.path, self.spec)])
        joined_cmd = ' '.join(cmd)
        try:
            proc = subprocess.Popen(joined_cmd, shell=True,
                                    universal_newlines=True,
                                    stdout=subprocess.PIPE,
                                    stderr=subprocess.PIPE)
            output, err = proc.communicate()
        except Exception as e:
            if err:
                self.log.debug('Errors occurred while running following command to get N-V-R-E:')
                self.log.debug(joined_cmd)
                self.log.error(err)
            raise rpkgError('Could not query n-v-r of %s: %s'
                            % (self.module_name, e))
        if err:
            self.log.debug('Errors occurred while running following command to get N-V-R-E:')
            self.log.debug(joined_cmd)
            self.log.error(err)
        if proc.returncode > 0:
            raise rpkgError('Could not get n-v-r-e from %s'
                            % os.path.join(self.path, self.spec))

        # Get just the output, then split it by ??, grab the first and split
        # again to get ver and rel
        first_line_output = output.split('??')[1]
        parts = first_line_output.split()
        if len(parts) != 4:
            raise rpkgError('Could not get n-v-r-e from %r'
                            % first_line_output)
        (self._module_name_spec,
         self._epoch,
         self._ver,
         self._rel) = parts

        # Most packages don't include a "Epoch: 0" line, in which case RPM
        # returns '(none)'
        if self._epoch == "(none)":
            self._epoch = "0"

    @property
    def nvr(self):
        """This property ensures the nvr attribute"""

        if not self._nvr:
            self.load_nvr()
        return self._nvr

    def load_nvr(self):
        """This sets the nvr attribute"""

        self._nvr = '%s-%s-%s' % (self.module_name, self.ver, self.rel)

    def load_repo_spec(self):
        if not os.path.isdir(self._specloc):
            self._repo_spec = os.path.abspath(self._specloc)
            return

        files = os.listdir(self._specloc)
        spec = None

        for f in files:
            if f.endswith('.spec') and not f.startswith('.'):
                spec = os.path.abspath(os.path.join(self._specloc, f))

        if self.preprocess:
            for f in files:
                if f.endswith('.spec.rpkg') and not f.startswith('.'):
                    spec = os.path.abspath(os.path.join(self._specloc, f))

        if not spec:
            raise rpkgError('No spec file found at {}.'
                            .format(self._specloc))

        self._repo_spec = spec

    def load_spec(self):
        """This sets the spec attribute"""
        self._spec = self.preprocess_spec(self.repo_spec)
        if not self.dump_spec_only:
            self.setup_source_symlinks()

    @cached_property
    def lookasidecache(self):
        return CGILookasideCache(
            self.lookasidehash, str(self.lookaside), str(self.lookaside_cgi),
            client_cert=self.cert_file, ca_cert=self.ca_cert)

    @property
    def ns_module_name(self):
        if not self._ns_module_name:
            self.load_ns_module_name()
        return self._ns_module_name

    @ns_module_name.setter
    def ns_module_name(self, ns_module_name):
        self._ns_module_name = ns_module_name

    def load_ns_module_name(self):
        """
        Loads the namespaced module name.
        """
        try:
            push_url = self.push_url
        except rpkgError:
            raise rpkgError('Unable to find push URL. Use --module-name.')

        replacements = {'user': self.user, 'module': '(.*)/?'}
        gitbaseurl_pattern = self.gitbaseurl%replacements + '$'
        anongiturl_pattern = self.anongiturl%replacements + '$'

        match = re.match(gitbaseurl_pattern, push_url)
        if not match:
            match = re.match(anongiturl_pattern, push_url)

        if match:
            ns_module_name = match.group(1)
        else:
            ns_module_name = urlparse(push_url).path.rsplit(':')[-1].strip('/')

        if ns_module_name.endswith('.git'):
            ns_module_name = ns_module_name[:-len('.git')]

        self._ns_module_name = ns_module_name

    def sources(self, outdir=None):
        """Download source files"""
        if not os.path.exists(self.sources_filename):
            return

        # Default to putting the files where the module is
        if not outdir:
            outdir = self.path

        sourcesf = SourcesFile(
            self.sources_filename, self.source_entry_type)

        for entry in sourcesf.entries:
            outfile = os.path.join(outdir, entry.file)
            self.lookasidecache.download(
                self.ns_module_name,
                entry.file, entry.hash, outfile,
                hashtype=entry.hashtype)

    def srpm(self, target=None, bcond_with=[], bcond_without=[]):
        """Create an srpm using hashtype from content in the module

        Requires sources already downloaded.
        """
        srpm_path = os.path.join(self.outdir, "%s-%s-%s.src.rpm"
                                 % (self.module_name, self.ver, self.rel))

        # See if we need to build the srpm
        if os.path.exists(srpm_path):
            self.log.debug('Srpm found, rewriting it.')

        cmd = ['rpmbuild']
        cmd.extend(self.rpmdefines)

        if target:
            cmd.extend(['--target', target])
        if self.quiet:
            cmd.append('--quiet')

        for arg in bcond_with:
            cmd.extend(['--with', arg])

        for arg in bcond_without:
            cmd.extend(['--without', arg])

        cmd.extend(['--nodeps', '-bs', self.spec])
        self._run_command(cmd, shell=True)
        return srpm_path

    def lint(self, info=False, rpmlintconf=None):
        """Run rpmlint over a built srpm

        Log the output and returns nothing
        rpmlintconf is the name of the config file passed to rpmlint if
        specified by the command line argument.
        """
        self.log.info('Working on {}...'.format(self.outdir))

        cmd = ['rpmlint']
        if info:
            cmd.extend(['-i'])
        if rpmlintconf:
            cmd.extend(["-f", os.path.join(self.path, rpmlintconf)])
        elif os.path.exists(os.path.join(self.path, ".rpmlint")):
            cmd.extend(["-f", os.path.join(self.path, ".rpmlint")])

        cmd.extend([self.outdir])

        # Run the command
        self._run_command(cmd, shell=True)

    def is_unpacked(self, dirpath, rpm_sources):
        """
        @deprecated

        Decide, whether we are dealing with "unpacked"
        or "packed" type of source content at the
        given dirpath.

        "packed": does not contain anything else
                  except ignored files or contains
                  at least one source referenced
                  by the given specfile

        "unpacked": is not "packed", meaning that
                    it contains at least one non-ignored
                    file and contains no file referenced
                    by the given specfile as a source

        "source" in these definitions is a filename
        specified in a Source or Patch .spec directive

        NOTE:

        This criterion needs source list parsed from
        a specfile to make the decision, which makes
        it potentially dependant on the given environment
        where it is executed. You can try to avoid this
        dependency e.g. by not using Patch and Source
        spec directives inside conditionals and not
        using environment-dependant rpmmacros in Patch
        and Source definitions.

        :param str dirpath: filesystem path to a directory
                with the source content
        :param list rpm_sources: list of tuples describing
                rpm sources

        :returns True if the directory content is of the
                unpacked type, False otherwise
        """
        for (filepath, num, flags) in rpm_sources:
            filename = os.path.basename(filepath)

            local_filepath = os.path.join(dirpath, filename)
            if os.path.isfile(local_filepath):
                return False

        ignore_file_regex = (
            '(^LICENSE|^README|^tito.props$|.spec$|^\.|^sources$)')
        ignored_file_filter = lambda f: not re.search(
            ignore_file_regex, f, re.IGNORECASE)

        if not list(filter(ignored_file_filter, os.listdir(self.path))):
            return False

        return True

    def make_source(self):
        """
        @deprecated

        Create source mentioned according to Source0 spec
        directive from an unpacked repository. Does nothing
        on a packed repo.

        NOTE:

        This method calls rpm's parseSpec, evaluation
        of which is heavily dependant on the environment
        (e.g. currently defined macros under /usr/lib/rpm
        and ~/.rpmmacros), where it is being executed.

        Therefore, it should be called in a clean environment
        of the target rpm distribution and architecture.

        Note that if you invoke this method directly on your
        host system, you need to trust the provided spec file.
        When specfile is parsed, %() constructs get evaluated
        and any system command can be executed from there.
        The behaviour is the same as if you called rpmbuild
        directly in your system.

        :returns path to the packed archive (alias Source0)
        """
        ts = rpm.ts()
        try:
            rpm.addMacro("_sourcedir", self.outdir)
            rpm_spec = ts.parseSpec(self.spec)
        except ValueError as e:
            raise RpmSpecParseException(str(e))

        if not self.is_unpacked(self.outdir, rpm_spec.sources):
            raise NotUnpackedException("Not an unpacked content.")

        source_zero_name = utils.find_source_zero(rpm_spec.sources)
        if not source_zero_name:
            raise NoSourceZeroException("Source zero not found")

        target_source_path = os.path.join(self.outdir, source_zero_name)

        name = rpm.expandMacro("%{name}")
        version = rpm.expandMacro("%{version}")
        rpm.reloadConfig()

        packed_dir_name = name + '-' + version
        utils.pack_sources(
            self.path,
            target_source_path,
            packed_dir_name
        )
        self.log.info('Wrote: {}'.format(target_source_path))
        return target_source_path

    def setup_source_symlinks(self):
        ts = rpm.ts()
        try:
            rpm_spec = ts.parseSpec(self.spec)
            rpm.reloadConfig()
        except ValueError as e:
            raise RpmSpecParseException(str(e))

        for (filepath, num, flags) in rpm_spec.sources:
            filename = os.path.basename(filepath)
            symlink_source = os.path.join(self.path, filename)
            symlink_dest = os.path.join(self.outdir, filename)

            if os.path.isfile(symlink_source) and not \
                    os.path.isfile(symlink_dest):
                os.symlink(symlink_source, symlink_dest)

    def preprocess_spec(self, spec_templ):
        out_spec_basename = os.path.basename(
            spec_templ[:-5] if spec_templ.endswith('.rpkg') else spec_templ)

        out_spec_path = os.path.join(self.outdir, out_spec_basename)

        if self.preprocess:
            preprocess(self.path, spec_templ, out_spec_path,
                       self.log, self.dump_spec_only, self.version_bump)
        else:
            shutil.copy(spec_templ, out_spec_path)

        self.log.info('Wrote: {}'.format(out_spec_path))
        return out_spec_path

    @property
    def module_name(self):
        """This property ensures the name attribute"""
        if not self._module_name_spec:
            self.load_nameverrel()
        return self._module_name_spec

    def add_tag(self, tagname=None, skip_edit=False):
        """Add a git tag to the repository

        Takes an optional tagname.

        Runs the tag command and returns nothing.
        """
        tagname = tagname or self.nvr

        if not re.match(r'^.*-[^-]+-[^-]+$', tagname):
            raise rpkgError("Tag {} does not have n-v-r format.".format(tagname))

        if tagname in self.repo.tags:
            raise rpkgError("Tag {} already exists.".format(tagname))

        cmd = ['git', 'tag', '-a', '-m',
               self.get_tag_message(tagname, skip_edit),
               tagname]

        utils.run_command(cmd, cwd=self.path)
        self.log.info('Created tag: %s', tagname)

        self.version_bump = False
        self.load_spec()

    def get_tag_message(self, new_tagname, skip_edit=False):
        tag_pattern = '{}*'.format(self.module_name)

        tagnames = self.repo.git.tag(
            '--list', '--sort=-taggerdate:unix', tag_pattern,
            '--merged').split()

        last_matching_tagname = None
        for tagname in tagnames:
            expr = r'^{}-[^-]+-[^-]+$'.format(self.module_name)
            if re.match(expr, tagname):
                last_matching_tagname = tagname
                break

        if last_matching_tagname:
            rev_list_expr = '{}..HEAD'.format(
                last_matching_tagname)
        else:
            rev_list_expr = 'HEAD'

        entries = [utils.get_changelog_entry(commit.message) for commit in
                   self.repo.iter_commits(rev_list_expr, paths=[self.path])]

        editor = self.repo.git.var('GIT_EDITOR')
        subject = 'Commits: {0}'.format(rev_list_expr)

        help_text = ("#\n"
        "# Write a changelog for tag:\n"
        "#   " + new_tagname + "\n"
        "# Lines starting with '#' will be ignored.\n")

        body = '\n'.join(entries) + '\n' + help_text
        if not skip_edit:
            body = utils.edit(editor, body)

        return '{0}\n\n{1}'.format(subject, body)

    def push(self, force=False):
        """Push changes to the remote repository"""
        self.check_repo(is_dirty=False, all_pushed=False)

        # see if our branch is tracking anything
        try:
            self.load_branch_merge()
        except Exception:
            self.log.warning('Current branch cannot be pushed anywhere!')

        untracked_patches = self.find_untracked_patches()
        if untracked_patches:
            self.log.warning(
                'Patches %s %s not tracked within either git or sources',
                ', '.join(untracked_patches),
                'is' if len(untracked_patches) == 1 else 'are')

        cmd = ['git', 'push', '--follow-tags']
        if self.quiet:
            cmd.append('-q')
        self._run_command(cmd, cwd=self.path)

    def check_repo(self, is_dirty=True, all_pushed=True):
        """Check various status of current repository

        :param bool is_dirty: Default to True. To check whether there is uncommitted changes.
        :param bool all_pushed: Default to True. To check whether all changes are pushed.
        :raises rpkgError: if any unexpected status is detected. For example,
            if changes are not committed yet.
        """
        if is_dirty:
            if self.repo.is_dirty():
                raise rpkgError('%s has uncommitted changes.  Use git status '
                                'to see details' % self.path)
        if all_pushed:
            branch = self.repo.active_branch
            try:
                remote = self.repo.git.config('--get', 'branch.%s.remote' % branch)
                merge = self.repo.git.config('--get', 'branch.%s.merge' % branch).replace(
                    'refs/heads', remote)
            except git.GitCommandError:
                raise rpkgError('Branch {0} does not track remote branch.\n'
                                'Use the following command to fix that:\n'
                                '    git branch -u origin/REMOTE_BRANCH_NAME'.format(branch))
            if self.repo.git.rev_list('%s...%s' % (merge, branch)):
                raise rpkgError('There are unpushed changes in your repo')

    def install(self, target=None, short=False, nocheck=False,
                bcond_with=[], bcond_without=[]):
        """Run rpm -bi on a module

        optionally for a specific target, short-circuit it, or
        specify with/without options

        Logs the output and returns nothing
        """
        cmd = ['rpmbuild']
        cmd.extend(self.rpmdefines)

        if target:
            cmd.extend(['--target', target])
        if short:
            cmd.append('--short-circuit')
        if nocheck:
            cmd.append('--nocheck')
        if self.quiet:
            cmd.append('--quiet')

        for arg in bcond_with:
            cmd.extend(['--with', arg])

        for arg in bcond_without:
            cmd.extend(['--without', arg])

        cmd.extend(['-bi', os.path.join(self.path, self.spec)])
        self._run_command(cmd, shell=True)

    def compile(self, target=None, short=False, nocheck=False,
                bcond_with=[], bcond_without=[]):
        """Run rpmbuild -bc on a module

        optionally for a specific target, short-circuit it, or
        specify with/without options

        Logs the output and returns nothing
        """
        cmd = ['rpmbuild']
        cmd.extend(self.rpmdefines)

        if target:
            cmd.extend(['--target', target])
        if short:
            cmd.append('--short-circuit')
        if nocheck:
            cmd.append('--nocheck')
        if self.quiet:
            cmd.append('--quiet')

        for arg in bcond_with:
            cmd.extend(['--with', arg])

        for arg in bcond_without:
            cmd.extend(['--without', arg])

        cmd.extend(['-bc', os.path.join(self.path, self.spec)])
        self._run_command(cmd, shell=True)

    def prep(self, target=None, bcond_with=[], bcond_without=[]):
        """Run rpm -bp on a module

        optionally for a specific target, short-circuit it, or
        specify with/without options

        Logs the output and returns nothing
        """
        cmd = ['rpmbuild']
        cmd.extend(self.rpmdefines)

        if target:
            cmd.extend(['--target', target])
        if self.quiet:
            cmd.append('--quiet')

        for arg in bcond_with:
            cmd.extend(['--with', arg])

        for arg in bcond_without:
            cmd.extend(['--without', arg])

        cmd.extend(['--nodeps', '-bp', os.path.join(self.path, self.spec)])
        self._run_command(cmd, shell=True)

    def local(self, target=None, bcond_with=[], bcond_without=[]):
        """rpmbuild locally for given target.

        Writes output to a log file and logs it to the logger
        Returns nothing
        """
        cmd = ['rpmbuild']
        cmd.extend(self.rpmdefines)

        if target:
            cmd.extend(['--target', target])
        if self.quiet:
            cmd.append('--quiet')

        for arg in bcond_with:
            cmd.extend(['--with', arg])

        for arg in bcond_without:
            cmd.extend(['--without', arg])

        cmd.extend(['-ba', os.path.join(self.path, self.spec)])

        logfile = '.build-%s-%s.log' % (self.ver, self.rel)
        logfile_path = os.path.join(self.outdir, logfile)

        cmd = '%s | tee %s' % (' '.join(cmd), logfile_path)
        try:
            # pipestatus is for checking the first command
            # when zsh is used.
            subprocess.check_call(
                '%s; exit "${PIPESTATUS[0]} ${pipestatus[1]}"' % cmd,
                shell=True)
        except subprocess.CalledProcessError:
            raise rpkgError(cmd)

    def verify_files(self):
        """
        Run rpmbuild -bl on a module to verify the %files section.
        """
        self.log.info('Working on {}...'.format(self.outdir))

        # setup the rpm command
        cmd = ['rpmbuild']
        cmd.extend(self.rpmdefines)

        if self.quiet:
            cmd.append('--quiet')

        cmd.extend(['-bl', os.path.join(self.outdir, '*.spec')])
        self._run_command(cmd, shell=True)
