import logging
import os
import re
import shutil
import tarfile
import textwrap
import subprocess
import tempfile

from pyrpkg.errors import rpkgError

log = logging.getLogger("__main__")


def pack_sources(dir_to_pack, target_path, pack_dir_as):
    """
    Create a gzipped tar archive from the given directory.

    :param str dir_to_pack: directory to be packed
    :param str target_path: path to the resulting archive
    :param str pack_dir_as: packed directory name inside the archive
    """
    log.debug("Packing {} as {} into {}...".format(
        dir_to_pack, pack_dir_as, target_path))

    def exclude(tar_info):
        exclude_git_pattern = r'(/.git$|/.git/|/.gitignore$)'
        if re.search(exclude_git_pattern, tar_info.name):
            log.debug("Excluding {}".format(tar_info.name))
            return None
        return tar_info

    tarball = tarfile.open(target_path, 'w:gz')
    tarball.add(dir_to_pack, pack_dir_as, filter=exclude)
    tarball.close()


def find_source_zero(rpm_sources):
    """
    For the given list of rpm_sources,
    return filename of the Source0.

    :param list rpm_sources

    :returns str: filename of Source0 or None
    """
    for (filepath, num, flags) in rpm_sources:
        if num == 0 and flags == 1:
            return os.path.basename(filepath)
    return None


def get_changelog_entry(commit_message):
    entry = '- ' + commit_message.split('\n')[0] if commit_message else ''
    return '\n'.join(textwrap.wrap(re.sub(r'(\s+)%([^%])', r'\1%%\2', entry), 80))


def run_command(cmd, shell=False, env=None, cwd=None):
    try:
        output = subprocess.check_output(cmd, shell=shell, env=env, cwd=cwd)
    except (subprocess.CalledProcessError, OSError) as e:
        raise rpkgError(e)
    except KeyboardInterrupt:
        raise rpkgError('Command is terminated by user.')
    except Exception as e:
        raise rpkgError(e)
    return output.decode('utf-8')


def edit(editor, text):
    tmp_f = tempfile.NamedTemporaryFile('w', delete=False)
    tmp_f.write(text)
    tmp_f.close()
    subprocess.check_call(editor.split()+[tmp_f.name])
    tmp_f = open(tmp_f.name, 'r')
    result = tmp_f.read()
    tmp_f.close()
    os.unlink(tmp_f.name)
    return result
